"""
Function that ignite from main_window, Dialogs to read data files for data,
channels, datatype
"""

import os
import json
import re
import traceback
from pathlib import Path

from obspy.core import read as read_ms
from obspy.io.reftek.core import Reftek130Exception

from sohstationviewer.model.mseed.mseed import MSeed
from sohstationviewer.model.data_type_model import DataTypeModel
from sohstationviewer.database.extractData import signatureChannels
from sohstationviewer.controller.util import validateFile, displayTrackingInfo


def loadData(dataType, tracking_box, listOfDir, reqWFChans=[], reqSOHChans=[],
             readStart=None, readEnd=None):
    """
    Go through root dir and read all files in that dir and its subdirs
    :param dataType: str - type of data read
    :param tracking_box: QTextBrowser - widget to display tracking info
    :param listOfDir: [str,] - list of directories selected by users
    :param reqWFChans: [str,] - requested waveform channel list
    :param reqSOHChans: [str,] - requested soh channel list
    :param readStart: [float,] - start time of read data
    :param readEnd: [float,] - finish time of read data
    :return dataObject: DataTypeModel - object that keep the data read from
        listOfDir
    """
    dataObject = None
    for d in listOfDir:
        if dataObject is None:
            try:
                dataObject = DataTypeModel.create_data_object(
                    dataType, tracking_box, d,
                    reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                    readStart=readStart, readEnd=readEnd)
            except Exception:
                fmt = traceback.format_exc()
                msg = f"Dir {d} can't be read due to error: {str(fmt)}"
                displayTrackingInfo(tracking_box, msg, "Warning")

            # if dataObject.hasData():
            #     continue
            # If no data can be read from the first dir, throw exception
            # raise Exception("No data can be read from ", d)
        # TODO: will work with select more than one dir later
        # else:
        #     dataObject.readDir(d)

    # return dataObject.plottingData
    return dataObject


def readChannels(tracking_box, listOfDir):
    """
    Scan available channels (to be used in channel preferences dialog). Since
        channels for RT130 is hard code, this function won't be applied for it.
    :param tracking_box: QTextBrowser - widget to display tracking info
    :param listOfDir: [str,] - list of directories selected by users
    :return dataObject: DataTypeModel - object that keep the data read from
        listOfDir
    """
    dataObject = None
    for d in listOfDir:
        if dataObject is None:
            # dataObject = Reftek.Reftek(parent, d)
            # if dataObject.hasData():
            #     continue

            # dataObject = MSeed_Text(parent, d, readChanOnly=True)
            dataObject = MSeed(tracking_box, d, readChanOnly=True)
            if len(dataObject.channels) == 0:
                # If no data can be read from the first dir, throw exception
                raise Exception("No data can be read from ", d)
        else:
            dataObject.readDir(d, readChanOnly=True)
    return dataObject.channels


def detectDataType(tracking_box, listOfDir):
    """
    Detect data type for the given directories using getDataTypeFromFile
    :param tracking_box: QTextBrowser - widget to display tracking info
    :param listOfDir: [str,] - list of directories selected by users
    :return: None/str -
        + if there are more than one data types detected,
            return None with a warning message
        + if only Unknown data type detected,
            return None with a warning message
        + if data type found, return data_type,
    """

    sign_chan_dataType_dict = signatureChannels()

    dirDataTypeDict = {}
    for d in listOfDir:
        dataType = "Unknown"
        for path, subdirs, files in os.walk(d):
            for fileName in files:
                path2file = Path(path).joinpath(fileName)
                if not validateFile(path2file, fileName):
                    continue
                ret = getDataTypeFromFile(path2file, sign_chan_dataType_dict)
                if ret is not None:
                    dataType, chan = ret
                    break
            if dataType != "Unknown":
                break
        if dataType == "Unknown":
            dirDataTypeDict[d] = ("Unknown", '_')
        else:
            dirDataTypeDict[d] = (dataType, chan)
    dataTypeList = {d[0] for d in dirDataTypeDict.values()}
    if len(dataTypeList) > 1:
        dirDataTypeStr = json.dumps(dirDataTypeDict)
        dirDataTypeStr = re.sub(r'\{|\}|"', '', dirDataTypeStr)
        dirDataTypeStr = re.sub(r', ', '\n', dirDataTypeStr)
        msg = (f"There are more than one types of data detected:\n"
               f"{dirDataTypeStr}\n\n"
               f"Please have only data that related to each other.")
        displayTrackingInfo(tracking_box, msg, "error")
        return

    elif dataTypeList == {'Unknown'}:
        msg = ("There are no known data detected.\n"
               "Please select different folder(s).")
        displayTrackingInfo(tracking_box, msg, "error")
        return

    return list(dirDataTypeDict.values())[0][0]


def getDataTypeFromFile(path2file, sign_chan_data_type_dict):
    """
    + Try to read mseed data from given file
        if catch TypeError: no data type detected => return None
        if catch Reftek130Exception: data type => return data type RT130
        otherwise data type is mseed which includes: q330, pegasus, centaur
    + Continue to identify data type for a file by checking if the channel
    in that file is a unique channel of a data type.
    :param path2file: str - absolute path to processed file
    :param sign_chan_data_type_dict: {str: str} - dict of unique chan for data
        type
    :return: str, str - detected data type, channel from which data type is
        detected
    """
    try:
        stream = read_ms(path2file)
    except TypeError:
        return
    except Reftek130Exception:
        return 'RT130', '_'

    for trace in stream:
        chan = trace.stats['channel']
        if chan in sign_chan_data_type_dict.keys():
            return sign_chan_data_type_dict[chan], chan
