import os

from tempfile import mkdtemp
import shutil

from sohstationviewer.controller.util import displayTrackingInfo
from sohstationviewer.conf import constants


class WrongDataTypeError(Exception):
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class DataTypeModel():
    def __init__(self, trackingBox, folder, readChanOnly=False,
                 reqWFChans=[], reqSOHChans=[],
                 readStart=0, readEnd=constants.HIGHEST_INT,
                 *args, **kwargs):
        """
        Super class for different data type to process data from data files

        :param trackingBox: QTextBrowser - widget to display tracking info
        :param folder: str - path to the folder of data
        :param readChanOnly: bool - if True, only read for channel name
        :param reqWFChans: list of str - requested waveform channel list
        :param reqSOHChans: list of str - requested SOH channel list
        :param readStart: float - requested start time to read
        :param readEnd: float - requested end time to read
        """
        self.trackingBox = trackingBox
        self.dir = folder
        self.reqSOHChans = reqSOHChans
        self.reqWFChans = reqWFChans
        self.readChanOnly = readChanOnly
        self.readStart = readStart
        self.readEnd = readEnd

        """
        processingLog: [(message, type)] - record the progress of processing
        """
        self.processingLog = []

        """
        Log data: info from log channels, soh messages, text file in dict:
        {chan_id: list of log strings}
        'TEXT': is the chan_id given by sohview for text only file.
        """
        self.logData = {'TEXT': []}

        """
        waveformData: data of waveform in dict:
        {set_key: {
            'files_info': {
                chan_id - str: [{
                    'path2file': absolute path to file - str,
                    'file_name': name of file - str,
                    'nets': list of network names in file - list of str
                    'stats': list of station names in file - list of str
                    'net_stats': list of (network, station)s in file - list of
                        tuple of str
                    'chan_ids': list of channel names in file - list of str
                    'start_epoch': start time of file - float
                    'end_epoch': end time of file - float
                    'traces_total': length of data - int
                    'spr': sample rate - float
                    'read': file has been read or not - bool
                    }]
            }
            'read_data': {
                chan_id - str: {
                    'samplerate': Sample rate of the data - float
                    'tracesInfo': [{
                        'samplerate': Sample rate of the data - float
                        'start_tm_epoch': Start epoch time of the trace - float
                        'end_tm_epoch': End epoch time of the trace - float
                        'size': Size of data - int
                        'times_f': np.memmap's file path for real epoch time
                            - str
                        'data_f': np.memmap's file path for data - str
                        },]
                    'times': times that has been trimmed and down- sampled for
                        plotting - np.array of float,
                    'data': data that has been trimmed and down-sampled for
                        plotting - np.array of float/int,
                    'chan_db_info': the plotting parameters got from database
                        for this channel - dict,
                    ax_wf: axes to draw the channel in WaveformWidget, to
                        combine 2 channels in need - matplotlib.axes.Axes
                }
            }
        }
        (np.memmap is used instead of data for waveform because waveform data
        are too large and make system crashed when try to merge and process)
        """
        self.waveformData = {}

        """
        SOHData: data for SOH info in dict:
        {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.SOHData = {}

        """
        massPosData: data for mass position info in the format similar to
        SOHData but there are 2 axes for drawing a mass possition channel,
        one is 'ax' to display in PlottingWidget, the other is 'ax_wf' to
        display in WaveformWidget.
        Channel name for mseed mass position: VP*
        Channel name for reftek mass position: MP* (given by SOHView)
                {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes,
                ax_wf: axes to draw the channel in WaveformWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.massPosData = {}

        """
        dataTime: time range of data sets:
        {setKey: [earliestepoch, latestepoch]} - {str: [float, float],}
        """
        self.dataTime = {}

        """
        The given data may include more than one data set which is station_id
        in mseed or (unit_id, exp_no) in reftek. User are allow to choose which
        data set to be displayed
        selectedKey: str - key of the data set to be displayed
        """
        self.selectedKey = None

        """
        gaps: gaps info in dict:
        {set_key: [list of gaps]} - {str: [[float, float],],}
        """
        self.gaps = {}

        """
         tmpDir: str - dir to keep memmap files.
         Will be deleted when object is deleted
        """
        self.tmpDir = mkdtemp()
        try:
            os.mkdir(self.tmpDir)
        except FileExistsError:
            shutil.rmtree(self.tmpDir)
            os.mkdir(self.tmpDir)

    def __del__(self):
        print("delete dataType Object")
        try:
            shutil.rmtree(self.tmpDir)
        except OSError as e:
            self.trackInfo(
                "Error deleting %s : %s" % (self.tmpDir, e.strerror),
                "error")
            print("Error deleting %s : %s" % (self.tmpDir, e.strerror))
        print("finish deleting")

    def hasData(self):
        """
        :return bool - True if there is any data can be read.
                       False if there is no valid data
        """
        if (len(self.logData) == 0 and len(self.SOHData) == 0 and
                len(self.massPosData) == 0 and len(self.waveformData) == 0):
            return False
        return True

    def trackInfo(self, text, type):
        """
        Display tracking info in tracking_box.
        Add all errors/warnings to processing_log.
        :param text: str - message to display
        :param type: str - type of message (error/warning/info)
        """
        displayTrackingInfo(self.trackingBox, text, type)
        if type != 'info':
            self.processingLog.append((text, type))

    @classmethod
    def create_data_object(cls, data_type, tracking_box, folder,
                           readChanOnly=False, reqWFChans=[], reqSOHChans=[],
                           readStart=0, readEnd=constants.HIGHEST_INT):
        if data_type == 'RT130':
            from sohstationviewer.model.reftek.reftek import RT130
            dataObject = RT130(
                tracking_box, folder, readChanOnly=readChanOnly,
                reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                readStart=readStart, readEnd=readEnd)
        else:
            from sohstationviewer.model.mseed.mseed import MSeed
            dataObject = MSeed(
                tracking_box, folder, readChanOnly=readChanOnly,
                reqWFChans=reqWFChans, reqSOHChans=reqSOHChans,
                readStart=readStart, readEnd=readEnd)
        return dataObject
