import pathlib
import os
from pathlib import Path
from datetime import datetime
from copy import deepcopy
from PySide2 import QtCore, QtWidgets

from sohstationviewer.view.ui.main_ui import UIMainWindow
from sohstationviewer.view.calendar.calendar_dialog import (
    CalendarDialog)
from sohstationviewer.view.file_list_widget import FileListItem
from sohstationviewer.view.plotting.waveform_dialog import WaveformDialog
from sohstationviewer.view.plotting.time_power_squared_dialog import (
    TimePowerSquaredDialog)

from sohstationviewer.view.db_config.data_type_dialog import DataTypeDialog
from sohstationviewer.view.db_config.param_dialog import ParamDialog
from sohstationviewer.view.db_config.channel_dialog import ChannelDialog
from sohstationviewer.view.db_config.plot_type_dialog import PlotTypeDialog
from sohstationviewer.view.channel_prefer_dialog import (
    ChannelPreferDialog)

from sohstationviewer.controller.processing import loadData, detectDataType

from sohstationviewer.database.proccessDB import executeDB_dict

from sohstationviewer.conf.constants import TM_FORMAT


class MainWindow(QtWidgets.QMainWindow, UIMainWindow):

    current_directory_changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui(self)
        """
        dir_names: [PosixPath,] - list of absolute path of data set directory
        """
        self.dir_names = []
        """
        data_type: str - type of data set
        """
        self.data_type = 'Unknown'
        """
        req_soh_chans: [str,] - list of State-Of-Health channels to read data
            from. For Reftek, the list of channels is fixed => may not need
        """
        self.req_soh_chans = []
        """
        req_wf_chans: [str,] - list of waveform channels to read data from.
             For Reftek, it is [int,] which is list of index of data stream
             to read data from.
        """
        self.req_wf_chans = []
        """
        start_tm: float - start time of data to read
        end_tm: float - end time of data to read
        """
        self.start_tm = 0
        self.end_tm = 0
        """
        data_object: DataTypeModel - Object that keep data read from data set
            for plotting
        """
        self.data_object = None
        """
        min_gap: float - minimum minutes of gap length to be display on gap bar
        """
        self.min_gap = None
        """
        ids_name: str - name of selected preferred channels list
        """
        self.ids_name = ''
        """
        ids: [str,] - selected preferred channels
        """
        self.ids = []
        """
        ids_data_type: str - data type of the preferred channels list
        """
        self.ids_data_type = 'Unknown'
        # Options
        """
        date_format: str - format for date
        """
        self.date_format = 'YYYY-MM-DD'
        """
        mass_pos_volt_range_opt: str - option for map value/color of
            mass position
        """
        self.mass_pos_volt_range_opt = 'regular'
        """
        bit_weight_opt: str - option for bitweight
        """
        self.bit_weight_opt = ''         # currently only need one option
        self.get_channel_prefer()
        self.yyyy_mm_dd_action.triggered.emit()

        """
        waveform_dlg: PlottingWidget - widget to display waveform channels'
            plotting
        """
        self.waveform_dlg = WaveformDialog(self)
        """
        tps_dlg: PlottingWidget - widget to display time-power-squared of
            waveform channels
        """
        self.tps_dlg = TimePowerSquaredDialog(self)

    @QtCore.Slot()
    def open_data_type(self):
        """
        Open a dialog to add/edit data types in DB
        """
        win = DataTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def open_param(self):
        """
        Open a dialog to add/edit parameters in DB
        """
        win = ParamDialog(self)
        win.show()

    @QtCore.Slot()
    def open_channel(self):
        """
        Open a dialog to add/edit channels in DB
        """
        win = ChannelDialog(self)
        win.show()

    @QtCore.Slot()
    def open_plot_type(self):
        """
        Open a dialog to view plot types and their description
        """
        win = PlotTypeDialog(self)
        win.show()

    @QtCore.Slot()
    def open_calendar(self):
        """
        Open calendar tool
        """
        calendar = CalendarDialog(self)
        calendar.show()

    @QtCore.Slot()
    def open_channel_preferences(self):
        """
        Open a dialog to view, select, add, edit, scan for preferred channels
            list.
        """
        dir_names = [os.path.join(self.cwd_line_edit.text(), item.text())
                     for item in self.open_files_list.selectedItems()]
        if dir_names == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return
        win = ChannelPreferDialog(self, dir_names)
        win.show()

    @QtCore.Slot()
    def all_chan_clicked(self):
        """
        When "All SOH" is checked,
            + If checked, clear current IDs textbox
            + If unchecked, set current IDs textbox if there is a preferred
                channels list selected. If no list selected, re-check "All SOH"
        """
        if not self.all_soh_chan_check_box.isChecked():
            if self.ids == []:
                self.all_soh_chan_check_box.setChecked(True)
            else:
                self.curr_soh_ids_name_line_edit.setText(self.ids_name)
        else:
            self.curr_soh_ids_name_line_edit.setText('')

    @QtCore.Slot()
    def set_date_format(self, display_format):
        """
        Sets the calendar format used by the QDateEdit text boxes.
        :param display_format: str - A valid display format to be used for date
            conversion.
        """
        self.time_to_date_edit.setDisplayFormat(display_format)
        self.time_from_date_edit.setDisplayFormat(display_format)
        self.date_format = display_format

    @QtCore.Slot()
    def open_files_list_item_double_clicked(self, item):
        """
        Handles the double-click event emitted when a user double-clicks on an
        item contained within openFilesList

        :param item: FileListItem - The item contained within openFilesList
            which was clicked.
        """
        print(f'Opening {item.text()}')

        # TODO: Do something with the Path object,
        # i.e., path.open(), or path.iterdir() ...

    @QtCore.Slot()
    def change_current_directory(self):
        """
        Constructs a QFileDialog to select a new working directory from which
            the user can load data. The starting directory is taken from
            cwdLineEdit.
        """
        fd = QtWidgets.QFileDialog(self)
        fd.setFileMode(QtWidgets.QFileDialog.Directory)
        fd.setDirectory(self.cwd_line_edit.text())
        fd.exec_()
        new_path = fd.selectedFiles()[0]
        self.set_current_directory(new_path)

    @QtCore.Slot()
    def read_selected_files(self):
        """
        Read data from selected files/directories, process and plot channels
            read from those according to current options set on the GUI
        """
        try:
            del self.data_object
        except AttributeError:
            pass

        self.req_soh_chans = (self.ids
                              if not self.all_soh_chan_check_box.isChecked()
                              else [])

        self.dir_names = [Path(self.cwd_line_edit.text()).joinpath(item.text())
                          for item in self.open_files_list.selectedItems()]
        if self.dir_names == []:
            msg = "No directories has been selected."
            QtWidgets.QMessageBox.warning(self, "Select directory", msg)
            return
        self.data_type = detectDataType(self.tracking_info_text_browser,
                                        self.dir_names)
        if self.data_type is None:
            return

        # get req_soh_chans
        if (not self.all_soh_chan_check_box.isChecked() and
                self.data_type != self.ids_data_type):
            msg = (f"DataType detected for the selected data set is "
                   f"{self.data_type} which is different to IDs' "
                   f"{self.ids_data_type}.\n"
                   f"SOHStationViewer will read all data available.\n"
                   f"Do you want to continue?")
            result = QtWidgets.QMessageBox.question(
                self, "Confirmation", msg,
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if result == QtWidgets.QMessageBox.No:
                return
            self.all_soh_chan_check_box.setChecked(True)
            self.curr_soh_ids_name_line_edit.setText('')
            self.req_soh_chans = []
        self.req_wf_chans = []
        if self.data_type == 'RT130':
            req_dss = []
            for idx, DSCheckbox in enumerate(self.ds_check_boxes):
                if DSCheckbox.isChecked():
                    req_dss.append(idx + 1)
            self.req_wf_chans = req_dss
        else:
            if self.wf_all_check_box.isChecked():
                self.req_wf_chans = ['*']
            elif self.wf_chans_line_edit.text().strip() != "":
                self.req_wf_chans = self.wf_chans_line_edit.text().split(",")
        start_tm_str = self.time_from_date_edit.date().toString(
            QtCore.Qt.ISODate)
        end_tm_str = self.time_to_date_edit.date().toString(QtCore.Qt.ISODate)
        self.start_tm = datetime.strptime(start_tm_str, TM_FORMAT).timestamp()
        self.end_tm = datetime.strptime(end_tm_str, TM_FORMAT).timestamp()
        self.data_object = loadData(self.data_type,
                                    self.tracking_info_text_browser,
                                    self.dir_names,
                                    reqWFChans=self.req_wf_chans,
                                    reqSOHChans=self.req_soh_chans,
                                    readStart=self.start_tm,
                                    readEnd=self.end_tm)
        self.replot_loaded_data()

    @QtCore.Slot()
    def replot_loaded_data(self):
        """
        Plot using data from self.data_object with the current options set
            from GUI
        """
        if self.detect_gap_check_box.isChecked():
            self.min_gap = self.gap_len_line_edit.text()
        else:
            self.min_gap = None
        do = self.data_object

        time_tick_total = 5  # TODO: let user choose max ticks to be displayed

        sel_key = do.selectedKey
        soh_data = deepcopy(do.SOHData[sel_key])
        soh_chans = list(soh_data.keys())
        mp_data = deepcopy(do.massPosData[sel_key])
        if len(self.req_wf_chans) != 0:
            wf_data = deepcopy(do.waveformData[sel_key]['readData'])
        else:
            wf_data = []
        self.plotting_widget.plot_channels(
            self.start_tm, self.end_tm, sel_key,
            do.dataTime[sel_key], soh_chans, time_tick_total,
            soh_data, mp_data, do.gaps[sel_key])

        peer_plotting_widgets = [self.plotting_widget]

        if self.tps_check_box.isChecked():
            peer_plotting_widgets.append(self.tps_dlg.plotting_widget)
            self.tps_dlg.set_data(
                self.data_type, ','.join([str(d) for d in self.dir_names]))
            self.tps_dlg.show()
            self.tps_dlg.plotting_widget.plot_channels(
                self.start_tm, self.end_tm, sel_key,
                do.dataTime[sel_key],
                wf_data)
        else:
            self.tps_dlg.hide()

        if self.req_wf_chans != []:
            # waveformPlot
            peer_plotting_widgets.append(self.waveform_dlg.plotting_widget)
            self.waveform_dlg.set_data(
                self.data_type, ','.join([str(d) for d in self.dir_names]))
            self.waveform_dlg.show()
            self.waveform_dlg.plotting_widget.plot_channels(
                self.start_tm, self.end_tm, sel_key,
                do.dataTime[sel_key], time_tick_total,
                wf_data, mp_data)
        else:
            self.waveform_dlg.hide()

        self.plotting_widget.set_peer_plotting_widgets(peer_plotting_widgets)
        self.waveform_dlg.plotting_widget.set_peer_plotting_widgets(
            peer_plotting_widgets)
        self.tps_dlg.plotting_widget.set_peer_plotting_widgets(
            peer_plotting_widgets)

    def set_current_directory(self, path=''):
        """
        Set all directories under current directory to self.open_files_list
        :param path: str - absolute path to current directory
        """
        # Remove entries when cwd changes
        self.open_files_list.clear()
        # Signal cwd changed, and gather list of files in new cwd
        self.current_directory_changed.emit(path)
        for dent in pathlib.Path(path).iterdir():
            if not dent.is_dir() or dent.name.startswith('.'):
                continue

            self.open_files_list.addItem(FileListItem(dent))

    def get_channel_prefer(self):
        """
        Read the current preferred channel list from database to set ids_name,
            ids, self.ids_data_type
        """
        self.ids_name = ''
        self.ids = []
        self.data_type = 'Unknown'
        rows = executeDB_dict('SELECT name, IDs, dataType FROM ChannelPrefer '
                              'WHERE current=1')
        if len(rows) > 0:
            self.ids_name = rows[0]['name']
            self.ids = [t.strip() for t in rows[0]['IDs'].split(',')]
            self.ids_data_type = rows[0]['dataType']

    def resizeEvent(self, event):
        """
        When MainWindow is resized, its plotting_widget need to initialize
            its size to fit the viewport.

        :param event: QResizeEvent - resize event
        """
        self.plotting_widget.init_size()
