# class with all plotting functions
from sohstationviewer.controller.util import getVal
from sohstationviewer.controller.plottingData import getMassposValueColors

from sohstationviewer.view.util.color import clr
from sohstationviewer.conf import constants


class Plotting:
    """
    Class that includes different methods to plot channels on a figure.
    """
    def __init__(self, parent, plotting_axes, parent_params):
        """
        :param parent: PlottingWidget - widget to plot channels
        :param plotting_axes: PlottingAxes - widget that includes a figure
            and methods related to create axes
        :param parent_params: Object - object that includes needed parameters
        """
        super().__init__()
        self.parent = parent
        self.params = parent_params
        self.plotting_axes = plotting_axes

    def plot_none(self):
        """
        Plot with nothing needed to show rulers.
        :return ax: matplotlib.axes.Axes - axes of the empty plot
        """
        plot_h = 0.00001
        bw_plots_distance = 0.0001
        self.parent.plotting_bot -= plot_h + bw_plots_distance
        ax = self.plotting_axes.create_axes(
            self.parent.plotting_bot, plot_h, has_min_max_lines=False)
        ax.x = None
        ax.plot([0], [0], linestyle="")
        return ax

    def plot_multi_color_dots(self, c_data, chan_db_info, chan_id,
                              ax, linked_ax):
        """
        plot scattered dots with colors defined by valueColors in database:
            Ex: *:W  means everything with white color
            Ex: -1:_|0:R|2.3:Y|+2.3:G  means
                value <= -1  => not plot
                value <= 0   => plot with R color
                value <= 2.3 => plot with Y color
                value > 2.3  => plot with G color
        Color codes are defined in colorSettings and limitted in 'valColRE'
            in dbSettings.py

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.plotting_axes.get_height(chan_db_info['height'])
            ax = self.plotting_axes.create_axes(
                self.parent.plotting_bot, plot_h,
                has_min_max_lines=False)

        x = []
        prev_val = -constants.HIGHEST_INT

        if chan_db_info['valueColors'] in [None, 'None', '']:
            chan_db_info['valueColors'] = '*:W'
        value_colors = chan_db_info['valueColors'].split('|')
        for vc in value_colors:
            v, c = vc.split(':')
            val = getVal(v)
            if c == '_':
                prev_val = val
                continue

            if v.startswith('+'):
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if c_data['data'][i] > val]
            elif v == '*':
                points = c_data['times']
            else:
                points = [c_data['times'][i]
                          for i in range(len(c_data['data']))
                          if prev_val < c_data['data'][i] <= val]
            x += points

            ax.plot(points, len(points) * [0], linestyle="",
                    marker='s', markersize=0.5,
                    zorder=constants.Z_ORDER['DOT'],
                    color=clr[c], picker=True, pickradius=3)
            prev_val = val

        total_samples = len(x)

        x = sorted(x)
        self.plotting_axes.set_axes_info(
            ax, [total_samples], chan_db_info=chan_db_info,
            linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_up_down_dots(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot channel with 2 different values, one above, one under center line.
        Each value has corresponding color defined in valueColors in database.
        Ex: 1:Y|0:R  means
            value == 1 => plot above center line with Y color
            value == 0 => plot under center line with R color
        Color codes are defined in colorSettings

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.plotting_axes.get_height(chan_db_info['height'])
            ax = self.plotting_axes.create_axes(
                self.parent.plotting_bot, plot_h,
                has_min_max_lines=False)

        val_cols = chan_db_info['valueColors'].split('|')
        points_list = []
        colors = []
        for vc in val_cols:
            v, c = vc.split(':')
            val = getVal(v)

            points = [c_data['times'][i]
                      for i in range(len(c_data['data']))
                      if c_data['data'][i] == val]
            points_list.append(points)
            colors.append(c)

        # down dots
        ax.plot(points_list[0], len(points_list[0]) * [-0.5], linestyle="",
                marker='s', markersize=2, zorder=constants.Z_ORDER['DOT'],
                color=clr[colors[0]], picker=True, pickradius=3)
        # up dots
        ax.plot(points_list[1], len(points_list[1]) * [0.5], linestyle="",
                marker='s', markersize=2, zorder=constants.Z_ORDER['DOT'],
                color=clr[colors[1]], picker=True, pickradius=3)
        x = points_list[0] + points_list[1]
        x = sorted(x)
        ax.set_ylim(-2, 2)
        self.plotting_axes.set_axes_info(
            ax, [len(points_list[0]), len(points_list[1])],
            chan_db_info=chan_db_info, linked_ax=linked_ax)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_time_dots(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot times only
        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id]

        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.plotting_axes.get_height(chan_db_info['height'])
            ax = self.plotting_axes.create_axes(
                self.parent.plotting_bot, plot_h)

        color = 'W'
        if chan_db_info['valueColors'] not in [None, 'None', '']:
            color = chan_db_info['valueColors'].strip()
        x = c_data['times']
        self.plotting_axes.set_axes_info(
            ax, [len(x)], chan_db_info=chan_db_info, linked_ax=linked_ax)

        ax.myPlot = ax.plot(x, [0] * len(x), marker='s', markersize=1.5,
                            linestyle='', zorder=constants.Z_ORDER['LINE'],
                            color=clr[color], picker=True,
                            pickradius=3)
        if linked_ax is None:
            ax.x = x
        else:
            ax.linkedX = x
        return ax

    def plot_lines_dots(self, c_data, chan_db_info, chan_id,
                        ax, linked_ax, info=''):
        """
        Plot lines with dots at the data points. Colors of dot and lines are
        defined in valueColors in database.
        Ex: L:G|D:W  means
            Lines are plotted with color G
            Dots are plotted with color W
        If D is not defined, dots won't be displayed.
        If L is not defined, lines will be plotted with color G
        Color codes are defined in colorSettings

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            soh_data[key][chan_id] or DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id] for waveform data
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :param info: str - additional info to be displayed on sub-title under
            main-title
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if linked_ax is not None:
            ax = linked_ax
        if ax is None:
            plot_h = self.plotting_axes.get_height(chan_db_info['height'])
            ax = self.plotting_axes.create_axes(
                self.parent.plotting_bot, plot_h)

        x, y = c_data['times'], c_data['data']
        self.plotting_axes.set_axes_info(
            ax, [len(x)], chan_db_info=chan_db_info,
            info=info, y=y, linked_ax=linked_ax)
        colors = {}
        if chan_db_info['valueColors'] not in [None, 'None', '']:
            color_parts = chan_db_info['valueColors'].split('|')
            for cStr in color_parts:
                obj, c = cStr.split(':')
                colors[obj] = c

        l_color = 'G'
        has_dot = False
        if 'L' in colors:
            l_color = colors['L']
        if 'D' in colors:
            d_color = colors['D']
            has_dot = True

        if not has_dot:
            ax.myPlot = ax.plot(x, y,
                                linestyle='-', linewidth=0.7,
                                color=clr[l_color])
        else:
            ax.myPlot = ax.plot(x, y, marker='s', markersize=1.5,
                                linestyle='-', linewidth=0.7,
                                zorder=constants.Z_ORDER['LINE'],
                                color=clr[l_color],
                                markerfacecolor=clr[d_color],
                                picker=True, pickradius=3)
        if linked_ax is None:
            ax.x = x
            ax.y = y
        else:
            ax.linkedX = x
            ax.linkedY = y
        return ax

    def plot_lines_s_rate(self, c_data, chan_db_info, chan_id, ax, linked_ax):
        """
        Plot line only for waveform data channel (seismic data). Sample rate
        unit will be displayed

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if c_data['samplerate'] >= 1.0:
            info = "%dsps" % c_data['samplerate']
        else:
            info = "%gsps" % c_data['samplerate']
        return self.plot_lines_dots(c_data, chan_db_info, chan_id,
                                    ax, linked_ax, info=info)

    def plot_lines_mass_pos(self, c_data, chan_db_info, chan_id,
                            ax, linked_ax):
        """
        Plot multi-color dots with grey line for mass position channel.
        Use getMassposValueColors() to get value_colors map based on
            Menu - MP Coloring selected from Main Window.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to DataTypeModel.__init__.
            mass_pos_data[key][chan_id]
        :param chan_db_info: dict - info of channel from DB
        :param chan_id: str - name of channel
        :param ax: matplotlib.axes.Axes - axes to draw plot of channel
        :param linked_ax: matplotlib.axes.Axes/None - axes of another channel
            linked to this channel => both channels' will be plotted on the
            same axes
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        value_colors = getMassposValueColors(
            self.params.mass_pos_volt_range_opt, chan_id,
            self.parent.c_mode, self.parent.processing_log,
            retType='tupleList')

        if value_colors is None:
            return

        if ax is None:
            plot_h = self.plotting_axes.get_height(chan_db_info['height'])
            ax = self.plotting_axes.create_axes(
                self.parent.plotting_bot, plot_h)

        ax.x, ax.y = c_data['times'], c_data['data']
        self.plotting_axes.set_axes_info(
            ax, [len(ax.x)], chan_db_info=chan_db_info, y=ax.y)
        ax.myPlot = ax.plot(ax.x, ax.y,
                            linestyle='-', linewidth=0.7,
                            color=self.parent.display_color['sub_basic'],
                            zorder=constants.Z_ORDER['LINE'])[0]
        colors = [None] * len(ax.y)
        sizes = [0.5] * len(ax.y)
        for i in range(len(ax.y)):
            count = 0
            prev_v = 0
            for v, c in value_colors:
                if count < (len(value_colors) - 1):
                    if prev_v < abs(ax.y[i]) <= v:
                        colors[i] = clr[c]
                        break
                else:
                    colors[i] = clr[c]
                    break
                prev_v = v
                count += 1
        ax.scatter(ax.x, ax.y, marker='s', c=colors, s=sizes,
                   zorder=constants.Z_ORDER['DOT'])
        return ax
